"""
This file contains the database models
"""

from django.db.models import Model, CharField, DateTimeField, BooleanField, ForeignKey, CASCADE, IntegerField, SET_NULL, PROTECT, ManyToManyField
from django.contrib.postgres.fields import ArrayField, JSONField
from django.contrib.postgres.indexes import GinIndex
from enum import Enum
from django_json_widget.widgets import JSONEditorWidget

#
# Common
#

# abstract models

class BasicCommon(Model):
    # todo: we cannot use foreign keys in the array here, so we have to keep the Tags table up to date by trigger or so.
    # todo: we could switch to a manytomany field instead?
    tags = ArrayField(CharField(max_length=30), size=8, blank=True)
    created_at = DateTimeField(auto_now_add=True)
    updated_at = DateTimeField(auto_now=True)

    class Meta:
        abstract = True
        indexes = [GinIndex(fields=['tags'])]


class NamedCommon(BasicCommon):
    name = CharField(max_length=30)
    description = CharField(max_length=255)

    def __str__(self):
        return self.name

    class Meta:
        abstract = True


# concrete models

class Tags(Model):
    # todo: figure out how to keep this in sync with tags columns (->BasicCommon)
    # todo: Or remove this altogether without keeping track of tags?
    title = CharField(max_length=30)
    description = CharField(max_length=255)


#
# I/O
#

# choices

class AbstractChoice(Model):
    """
    Abstract class for all derived 'choices' models.
    We define a 'choice' as an item that you can pick from a predefined list.
    In the derived classes, we use an enum.Enum to define such a predefined list here in code.
    All values of the enums are then put automagically into the database in the populate module, which
    is/can_be/should_be called in the last migration step to populate the database with inital values
    for our 'static choices'.

    Design decision: Django also provides the 'choices' property on fields which sort of limits the number of choices
    one can make, and which sort of does some validation. In our opinion the validation is done in the wrong place, and
    no data consistency is enforced.
    So, we decided to follow Django's own hint, see https://docs.djangoproject.com/en/2.0/ref/models/fields/#choices
    "you’re probably better off using a proper database table with a ForeignKey"

    You can find the derived AbstractChoice classes being used as ForeignKey in other models, thus enforcing data
    consistency at database level.
    """
    value = CharField(max_length=30, editable=True, null=False, blank=False, unique=True, primary_key=True)

    class Meta:
        abstract = True

    def __str__(self):
        return self.value


class RoleChoice(AbstractChoice):
    """Defines the model and predefined list of possible RoleChoice's for WorkIORole.
    The items in the Choises class below are automagically populated into the database via a data migration."""
    class Choices(Enum):
        CORRELATOR = "correlator"
        BEAMFORMER = "beamformer"
        INSPECTION_PLOTS = "inspection plots"
        CALIBRATOR = "calibrator"
        TARGET = "target"
        INPUT_OUTPUT = "input, output"


class DatatypeChoice(AbstractChoice):
    """Defines the model and predefined list of possible DatatypeChoice's for WorkIORole.
    The items in the Choises class below are automagically populated into the database via a data migration."""
    class Choices(Enum):
        VISIBILITIES = "visibilities"
        TIME_SERIES = "time series"
        INSTRUMENT_MODEL = "instrument model"
        IMAGE = "image"
        QUALITY = "quality"

class DataformatChoice(AbstractChoice):
    """Defines the model and predefined list of possible DataformatChoice's for WorkRequestRelationDraft and WorkRequestRelationBlueprint.
    The items in the Choises class below are automagically populated into the database via a data migration."""
    class Choices(Enum):
        MEASUREMENTSET = "MeasurementSet"
        HDF5 = "HDF5"


class CopyReasonChoice(AbstractChoice):
    """Defines the model and predefined list of possible CopyReasonChoice's for RunDraft and WorkRequestDraft.
    The items in the Choises class below are automagically populated into the database via a data migration."""
    class Choices(Enum):
        TEMPLATE = "template"
        REPEATED = "repeated"


# concrete models

class WorkIORole(BasicCommon):
    role = ForeignKey('RoleChoice', null=False, on_delete=PROTECT)
    datatype = ForeignKey('DatatypeChoice', null=False, on_delete=PROTECT)
    dataformats = ManyToManyField('DataformatChoice', blank=True)
    outputs = ForeignKey("WorkRequestTemplate", related_name='role_output', on_delete=CASCADE)
    inputs = ForeignKey("WorkRequestTemplate", related_name='role_input', on_delete=CASCADE)


#
# Templates
#

# abstract models

class Template(NamedCommon):
    formfield_overrides = {
        JSONField: {'widget': JSONEditorWidget},
    }
    version = CharField(max_length=30)
    schema = JSONField()

    class Meta:
        abstract = True


# concrete models

class GeneratorTemplate(Template):
    create_func = CharField(max_length=30)


class DefaultGeneratorTemplate(BasicCommon):
    name = CharField(max_length=30, unique=True)
    template = ForeignKey("GeneratorTemplate", on_delete=PROTECT)


class RunTemplate(Template):
    pass


class DefaultRunTemplate(BasicCommon):
    name = CharField(max_length=30, unique=True)
    template = ForeignKey("RunTemplate", on_delete=PROTECT)


class WorkRequestTemplate(Template):
    validation_code_js = CharField(max_length=30)


class DefaultWorkRequestTemplate(BasicCommon):
    name = CharField(max_length=30, unique=True)
    template = ForeignKey("WorkRequestTemplate", on_delete=PROTECT)


class WorkRelationSelectionTemplate(Template):
    pass


class DefaultWorkRelationSelectionTemplate(BasicCommon):
    name = CharField(max_length=30, unique=True)
    template = ForeignKey("WorkRelationSelectionTemplate", on_delete=PROTECT)


#
# Instance Objects
#

class Cycle(NamedCommon):
    start = DateTimeField()


class Project(NamedCommon):
    # cycle is protected since we have to manually decide to clean up projects with a cycle or keep them without cycle
    cycle = ForeignKey('Cycle', related_name='projects', on_delete=PROTECT, null=True)
    priority = IntegerField(default=0)  # todo: define a value for the default priority
    can_trigger = BooleanField(default=False)
    private_data = BooleanField(default=True)


class RunSet(NamedCommon):
    generator_parameters_doc = JSONField()
    generator_template = ForeignKey('GeneratorTemplate', on_delete=SET_NULL, null=True)
    project = ForeignKey('Project', related_name="run_sets", on_delete=PROTECT) # protected to avoid accidents


class RunDraft(NamedCommon):
    requirements_doc = JSONField()
    copies = ForeignKey('RunDraft', related_name="copied_from", on_delete=SET_NULL, null=True)
    copy_reason = ForeignKey('CopyReasonChoice', null=False, on_delete=PROTECT)
    generator_param = CharField(max_length=200)  # todo: check if the field size is good enough
    run_set = ForeignKey(RunSet, related_name='run_drafts', on_delete=CASCADE)
    generator_source = ForeignKey(RunSet, on_delete=SET_NULL, null=True)
    template = ForeignKey(RunTemplate, on_delete=CASCADE)


class RunBlueprint(NamedCommon):
    requirements_doc = JSONField()
    do_cancel = BooleanField()
    template = ForeignKey('RunTemplate', on_delete=CASCADE)
    draft = ForeignKey('RunDraft', related_name='related_run_blueprint', on_delete=CASCADE)


class WorkRequestDraft(NamedCommon):
    requirements_doc = JSONField()
    copies = ForeignKey('WorkRequestDraft', related_name="copied_from", on_delete=SET_NULL, null=True)
    copy_reason = ForeignKey('CopyReasonChoice', null=False, on_delete=PROTECT)
    run_draft = ForeignKey(RunDraft, on_delete=CASCADE)
    template = ForeignKey('WorkRequestTemplate', on_delete=CASCADE)


class WorkRequestBlueprint(NamedCommon):
    requirements_doc = JSONField()
    do_cancel = BooleanField()
    template = ForeignKey('WorkRequestTemplate', on_delete=CASCADE)
    draft = ForeignKey('WorkRequestDraft', related_name='related_work_request_blueprint', on_delete=CASCADE)
    run_blueprint = ForeignKey('RunBlueprint', on_delete=CASCADE)


class WorkRequestRelationDraft(BasicCommon):
    selection_doc = JSONField()
    dataformat = ForeignKey('DataformatChoice', null=False, on_delete=PROTECT)
    producer = ForeignKey(WorkRequestDraft, related_name='produced_by', on_delete=CASCADE)
    consumer = ForeignKey(WorkRequestDraft, related_name='consumed_by', on_delete=CASCADE)
    input_role = ForeignKey(WorkIORole, related_name='inputs_work_request_relation_draft', on_delete=CASCADE)
    output_role = ForeignKey(WorkIORole, related_name='outputs_work_request_relation_draft', on_delete=CASCADE)
    selection_template = ForeignKey(WorkRelationSelectionTemplate, on_delete=CASCADE)


class WorkRequestRelationBlueprint(BasicCommon):
    selection_doc = JSONField()
    dataformat = ForeignKey('DataformatChoice', null=False, on_delete=PROTECT)
    producer = ForeignKey(WorkRequestBlueprint, related_name='produced_by', on_delete=CASCADE)
    consumer = ForeignKey(WorkRequestBlueprint, related_name='consumed_by', on_delete=CASCADE)
    input_role = ForeignKey(WorkIORole, related_name='inputs_work_request_relation_blueprint', on_delete=CASCADE)
    output_role = ForeignKey(WorkIORole, related_name='outputs_work_request_relation_blueprint', on_delete=CASCADE)
    draft = ForeignKey(WorkRequestRelationDraft, on_delete=CASCADE)
    selection_template = ForeignKey(WorkRelationSelectionTemplate, on_delete=CASCADE)

